// ------------------------------------------------------------------
// --- simple bootloader for ATmega128/ATmega8515                 ---
// ---                                                            ---
// --- XTAL: 11.0592 MHz                        RS232: 115k2 baud ---
// ------------------------------------------------------------------
// if you use a different XTAL frequency, change the BAUD rate values
// in serial.c and write your MHz value to delayms in delay.c in place
// of ldi r24,11 i.e. for a 8MHz XTAL write ldi r24,8
// ------------------------------------------------------------------
// ATmega128  fuses: C02F     lock: 2F     extended fuses: FF
//
// RWW section 0000-EFFF, NRWW section F000-FFFF
// for bootloader to work, .text section must be relocated to 
// 0x1E000 (2 * F000) in makefile, i.e. add
// ,--section-start=.text=0x1E000
// to LDFLAGS
// ------------------------------------------------------------------
// ATmega8515 fuses: C02F     lock: 2F
//
// RWW section 000-BFF, NRWW section C00-FFF
// for bootloader to work, .text section must be relocated to 
// 0x1800 (2 * C00) in makefile, i.e. add
// ,--section-start=.text=0x1800
// to LDFLAGS
// ------------------------------------------------------------------

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <avr/boot.h>
#include <util/crc16.h>
#include <avr/wdt.h>

#define STATE_EXIT	0
#define STATE_IDLE  1
#define STATE_WRITE 2

volatile uint8_t state;
volatile uint16_t bytecnt; 
volatile uint16_t pagecnt;

char flash_page_buf[SPM_PAGESIZE];

char* magic = "JeTaM";

// forward declarations
void serial_rx_int(uint8_t d);

#include "flash.c"
#include "serial.c"
#include "delay.c"

#ifdef _AVR_IOM128_H_
// 128 has 512 pages, 32 are for BL
#define NUM_APP_PAGES 480
#else
// 8515 has 128 pages, 32 are for BL
#define NUM_APP_PAGES 96
#endif

// ------------------------------------------------------------------

void hw_init(void)
{
	ACSR |= _BV(ACD);		// analog comparator poweroff
#ifdef _AVR_IOM128_H_	
	MCUCR = _BV(IVCE);		// relocate interrupt vectors to bootloader section
	MCUCR = _BV(IVSEL);
#else
#warning "Compiling for 8515"
	GICR = _BV(IVCE);		// relocate interrupt vectors to bootloader section
	GICR = _BV(IVSEL);
#endif
}

// ------------------------------------------------------------------

void hw_deinit(void)
{
	serial_disable();
	
#ifdef _AVR_IOM128_H_	
	MCUCR = _BV(IVCE);	// restore interrupt vectors back to application
	MCUCR = 0;
#else
	GICR = _BV(IVCE);	// restore interrupt vectors back to application
	GICR = 0;
#endif
	ACSR = 0;
}

// ------------------------------------------------------------------

// volatile uint16_t fcrc, fpused;

int main(void)
{
	cli();				// disable interrupts
	wdt_enable(WDTO_2S);// enable Watchdog with 2 sec timeout
	hw_init();			// general hw init
	serial_init();		// init serial interface
//	boot_lock_bits_set (_BV(BLB11));	// SPM is not allowed to write to the Bootloader section
	
	comsendstr_P(PSTR("Simple bootloader v1.0\n"));
	state = STATE_EXIT;
	
	sei();
	
	uint8_t i = 30;
	while(i--) {
		delayms(100);	// give user some time to press the magic sequence and enter STATE_IDLE
		wdt_reset();
	}
	
	while(STATE_EXIT != state) {
		wdt_reset();
		asm("nop");
	}
	
	// cleanup and exit bootloader
	cli();
	comsendstr_P(PSTR("Starting application\n"));
	delayms(10);
	serial_disable();
	hw_deinit();
	wdt_disable();	// first to be enabled, last to be disabled!
	asm("clr r30\n\t"
		"clr r31	\n\t"
		"ijmp		\n\t");
	return 0;
}

// ------------------------------------------------------------------

void serial_rx_int(uint8_t d)
{
	switch(state)
	{
		case STATE_EXIT:
		if (magic[bytecnt++] != d) {
			bytecnt = 0;
		}
		if (bytecnt == strlen(magic)) {
			state = STATE_IDLE;
			pagecnt = 0;
			comsendstr_P(PSTR("RDY\n"));
		}
		break;
		// --------------------------------------
		case STATE_IDLE:
		if (('W' == d)&&(pagecnt < NUM_APP_PAGES)) { 
			state = STATE_WRITE; 
			bytecnt = 0;
		} else
		if ('X' == d) {
			state = STATE_EXIT;
		} else {
			comsendstr_P(PSTR("ERR\n"));
			break;
		}
		comsendstr_P(PSTR("OK\n"));
		break;
		// --------------------------------------
		case STATE_WRITE:		// write page data command
			if (bytecnt < SPM_PAGESIZE) {
				flash_page_buf[bytecnt++] = d;
			}
			if (bytecnt == SPM_PAGESIZE) {
				flash_write_page(pagecnt, flash_page_buf);
				comsendhex(pagecnt);
				comsendchar('\n');
				comsendhex16(flash_page_crc(pagecnt));
				comsendstr_P(PSTR("\nOK\n"));
				pagecnt++;		// increase page counter
				state = STATE_IDLE;
			}
		break;
	}
}
